###############################################################################################
# SORTITION MAGIC
# Simulated annealing for random selection of citizens' assemblies
# Version 8.6, Release date: 11.01.2025
###############################################################################################

start_time_script <- Sys.time()

# Function to safely load packages
safe_load_packages <- function() {
  required_packages <- c("GenSA", "ggplot2", "ineq", "dplyr", "beepr", "openxlsx")
  
  for(pkg in required_packages) {
    if(!requireNamespace(pkg, quietly = TRUE)) {
      install.packages(pkg)
    }
    suppressWarnings(suppressMessages(library(pkg, character.only = TRUE)))
  }
}

# Enable byte code compilation for better performance
library(compiler)
enableJIT(3)

# Safely load required packages
safe_load_packages()

# Simple console output function with immediate flushing
console_and_file_print <- function(message) {
  cat(message, "\n")
  flush.console()
}

trimws_str <- function(x) {
  if(is.character(x)) trimws(x) else x
}

######################
# Settings variables #
######################

script.dir <- getSrcDirectory(function(x) {x})
main_directory <- '..'
input_file_directory <- "data_and_settings"
file_settings <- TRUE
settings_filename <- 'settings.xlsx'

# Set working directory and load settings
setwd(input_file_directory)
settings <- read.xlsx(settings_filename, sheet = 1, rowNames = TRUE)

# Extract settings
settings_list <- list(
  input_filename = as.character(settings["input_filename", "setting_value"]),
  par_filename = as.character(settings["par_filename", "setting_value"]),
  draw_name = as.character(settings["draw_name", "setting_value"]),
  assembly_size = as.numeric(settings["assembly_size", "setting_value"]),
  draws_number = as.numeric(settings["draws_number", "setting_value"]),
  SA_script_max_time = as.numeric(settings["SA_max_time", "setting_value"]),
  SA_temperature = as.numeric(settings["SA_temperature", "setting_value"]),
  SA_seed = as.numeric(settings["SA_seed", "setting_value"]),
  SA_max_iterations = as.numeric(settings["SA_max_iterations", "setting_value"]),
  SA_nb_stop_imp = as.numeric(settings["SA_nb_stop_imp", "setting_value"]),
  SA_threshold_stop = as.numeric(settings["SA_threshold_stop", "setting_value"]),
  SA_max_call = as.numeric(settings["SA_max_call", "setting_value"]),
  finish_signal = as.logical(as.numeric(settings["finish_signal", "setting_value"])),
  draws_files = as.logical(as.numeric(settings["draws_files", "setting_value"])),
  household_switch = as.logical(as.numeric(settings["household_duplicate", "setting_value"]))
)

# Set results filename
xlsx_results_filename <- paste0(
  "Results-", 
  settings_list$draw_name, 
  "-", 
  settings_list$draws_number, 
  if(settings_list$draws_number == 1) "-draw.xlsx" else "-draws.xlsx"
)

# Load and preprocess data with error handling
load_and_preprocess_data <- function(input_filename, par_filename) {
  tryCatch({
    # Load volunteers data
    volunteers <- read.xlsx(input_filename, sheet = 1, rowNames = TRUE)
    volunteers_names <- lapply(names(volunteers), trimws_str)
    volunteers <- data.frame(lapply(volunteers, trimws_str))
    names(volunteers) <- volunteers_names
    
    # Load characteristics data
    characteristics <- read.xlsx(par_filename, sheet = 1)
    characteristics_names <- lapply(names(characteristics), trimws_str)
    characteristics <- data.frame(lapply(characteristics, trimws_str))
    names(characteristics) <- characteristics_names
    
    # Process characteristics
    characteristics$counter <- rep(0, nrow(characteristics))
    characteristics$priority <- sapply(characteristics$priority, function(x) if(as.numeric(x) > 1) as.numeric(x) else 1)
    characteristics$category <- sapply(characteristics$category, function(x) gsub(" ", ".", x))
    
    list(
      volunteers = volunteers,
      characteristics = characteristics,
      INPUT_SIZE = nrow(volunteers),
      categories_number = nrow(characteristics)
    )
  }, error = function(e) {
    stop("Error loading data: ", e$message)
  })
}

# Load data
data <- load_and_preprocess_data(settings_list$input_filename, settings_list$par_filename)
volunteers <- data$volunteers
characteristics <- data$characteristics
INPUT_SIZE <- data$INPUT_SIZE
categories_number <- data$categories_number

# Pre-compute characteristic vectors for faster access
char_vectors <- list(
  category = c(characteristics$category),
  priority = as.numeric(c(characteristics$priority)),
  value = as.integer(c(characteristics$value)),
  feature = c(characteristics$feature)
)

# Create feature matrix for faster lookups
feature_matrix <- matrix(0, nrow = INPUT_SIZE, ncol = categories_number)
for(k in seq_len(categories_number)) {
  feature_matrix[, k] <- volunteers[[char_vectors$category[k]]] == char_vectors$feature[k]
}

# Optimized evaluation function
evaluation_function <- cmpfun(function(v, b, draw_no, draws_files) {
  # Pre-compute v indices once
  v_indices <- (round(v) %% INPUT_SIZE) + 1
  
  # Fast duplicate check
  if(length(unique(v_indices)) != length(v_indices)) {
    return(99999999)
  }
  
  # Vectorized household check
  if(settings_list$household_switch) {
    household_ids <- volunteers$HOUSEHOLD_ID[v_indices]
    if(length(unique(household_ids)) != length(v_indices)) {
      return(99999999)
    }
  }
  
  # Vectorized feature counting using pre-computed matrix
  counter <- colSums(feature_matrix[v_indices, , drop = FALSE])
  
  # Vectorized squared difference calculation
  ret <- sum(char_vectors$priority * (counter - char_vectors$value)^2)
  
  if(ret == 0 || b) {
    characteristics$counter <- characteristics$counter + counter
    characteristics_summary$counter <- characteristics_summary$counter + counter
    
    if(draws_files) {
      drawed_df <- data.frame(
        No = rownames(volunteers)[v_indices],
        ID = volunteers$ID[v_indices]
      )
      write.xlsx(drawed_df, 
                 paste0('result_', draw_no, '.xlsx'),
                 sheetName = paste0('result_', draw_no))
    }
  }
  
  ret
})

# Initialize results tracking
characteristics_summary <- characteristics
Nv <- 1:settings_list$assembly_size
draws_times <- data.frame()
function_value <- data.frame()
results_summary <- data.frame(
  'volunteers' = as.character(volunteers$ID),
  'drawn_counter' = rep(0, nrow(volunteers)),
  row.names = 'volunteers',
  fix.empty.names = FALSE
)

# Change directory for results
setwd(main_directory)
setwd("./results")

console_and_file_print(paste("Drawing started at:", format(Sys.time(), format = "%H:%M:%S")))
console_and_file_print(paste0(
  "Drawing panel(s) for: ",
  settings_list$draw_name,
  ", starting temperature: ",
  settings_list$SA_temperature
))

start_draws_time <- Sys.time()

# Main drawing loop
for(draw_no in 1:settings_list$draws_number) {
  console_and_file_print(paste("Starting draw no:", draw_no, 'of', settings_list$draws_number))
  
  tryCatch({
    characteristics$counter <- rep(0, categories_number)
    start_time <- Sys.time()
    
    # Initialize with random selection of N unique volunteers
    initial_params <- sample(1:INPUT_SIZE, settings_list$assembly_size, replace = FALSE)
    
    RESULT <- GenSA(
      par = as.numeric(initial_params),
      fn = evaluation_function,
      lower = as.numeric(rep(1, settings_list$assembly_size)),
      upper = as.numeric(rep(INPUT_SIZE, settings_list$assembly_size)),
      b = FALSE,
      draw_no = draw_no,
      draws_files = settings_list$draws_files,
      control = list(
        max.time = settings_list$SA_script_max_time,
        smooth = TRUE,
        verbose = FALSE,
        temperature = settings_list$SA_temperature,
        threshold.stop = settings_list$SA_threshold_stop,
        max.call = settings_list$SA_max_call,
        maxit = settings_list$SA_max_iterations,
        nb.stop.improvement = settings_list$SA_nb_stop_imp,
        seed = settings_list$SA_seed
      )
    )
    
    end_time <- Sys.time()
    
    # Process results
    new_time <- difftime(end_time, start_time, units = "secs")
    draws_times <- rbind(draws_times, round(as.numeric(new_time), 2))
    
    tracemat <- data.frame(RESULT$trace.mat)
    custom_min <- function(x) if(length(x) > 0) min(x) else Inf
    function_value <- rbind(function_value, custom_min(tracemat$current.minimum))
    
    # Update characteristics
    characteristics$counter <- rep(0, categories_number)
    evaluation_function(RESULT$par, TRUE, draw_no, settings_list$draws_files)
    
    # Update characteristics summary
    characteristics_summary[paste0('draw_', draw_no)] <- rep(0, categories_number)
    selected_indices <- (round(RESULT$par) %% INPUT_SIZE) + 1
    
    # Vectorized update of characteristics summary
    for(k in 1:categories_number) {
      input_category <- gsub(" ", ".", characteristics_summary[k, 'category'])
      matches <- characteristics_summary[k, 'feature'] == 
        volunteers[selected_indices, input_category]
      characteristics_summary[k, paste0('draw_', draw_no)] <- sum(matches)
      characteristics_summary[k, 'counter'] <- 
        characteristics_summary[k, 'counter'] + sum(matches)
    }
    
    # Update results summary
    results_summary[paste0('draw_', draw_no)] <- rep(0, nrow(volunteers))
    selected_ids <- volunteers$ID[selected_indices]
    for(id in selected_ids) {
      results_summary[as.character(id), paste0('draw_', draw_no)] <- 
        results_summary[as.character(id), paste0('draw_', draw_no)] + 1
      results_summary[as.character(id), 'drawn_counter'] <- 
        results_summary[as.character(id), 'drawn_counter'] + 1
    }
    
  }, error = function(e) {
    console_and_file_print(paste("Error in draw", draw_no, ":", e$message))
  }, finally = {
    console_and_file_print(paste("Drawing", draw_no, "done, function value:", RESULT$value))
  })
}

end_draws_time <- Sys.time()

# Results processing and saving
draws_times <- data.frame(t(draws_times))
function_value <- data.frame(t(function_value))

for (i in 1:ncol(draws_times)) {
  newColName <- paste('time_no', i, sep = "_")
  colnames(draws_times)[i] <- newColName
  colnames(function_value)[i] <- newColName
}

# Create and configure workbook
wb <- createWorkbook()

# Save results summary data
addWorksheet(wb, sheetName = paste('Results per volunteer'))
writeDataTable(wb, sheet = 1, results_summary, rowNames = TRUE)
setColWidths(wb, sheet = 1, cols = 1:(2+settings_list$draws_number), width = "auto")

results_summary <- results_summary[order(results_summary$drawn_counter),]
results_summary$ID <- rownames(results_summary)

# Check for identical columns
identical_columns <- duplicated(t(results_summary[,c(3:length(results_summary)-1)]))

# Gini Index calculation
gini_ind = round(Gini(results_summary$drawn_counter) * 100, 2)

# Calculate ten percent intervals
df_sorted <- results_summary
total_rows <- nrow(df_sorted)
ten_percent_rows <- round(0.1 * total_rows)
ten_percent_df <- data.frame(1:(ten_percent_rows+1))
ll <- split(df_sorted$drawn_counter, sort(rep_len(1:10, nrow(df_sorted))))
for (i in 1:10) {
  if (length(unlist(ll[i])) != (ten_percent_rows + 1)) {
    new_column <- c(unlist(ll[i]), rep(NA, (ten_percent_rows+1) - length(unlist(ll[i]))))
  } else {
    new_column <- c(unlist(ll[i]))
  }
  ten_percent_df <- cbind(ten_percent_df, new_column)
}
ten_percent_df <- ten_percent_df[, -1]
colnames(ten_percent_df) <- c('Top 10%', '2 - 10%', '3 - 10%', '4 - 10%', '5 - 10%', 
                              '6 - 10%', '7 - 10%', '8 - 10%', '9 - 10%', 'Bottom 10%')

# Total Equality Index calculation
draw_mean <- mean(df_sorted$drawn_counter, na.rm = TRUE)
interval_list <- list()
mean_list <- list()
for (i in 1:10) {
  aa <- mean(ten_percent_df[,11 - i], na.rm = TRUE)
  mean_list <- rbind(mean_list, round(aa, 2))
  tt <- abs(draw_mean - aa)
  interval_list <- rbind(interval_list, round(tt, 2))
}
interval_eq <- data.frame(colnames(ten_percent_df), mean_list, interval_list)
colnames(interval_eq) <- c('Intervals', 'Mean', 'Deviation')
rownames(interval_eq) <- c()
df_sorted$draw_dev <- abs(df_sorted$drawn_counter - draw_mean)
tot_eq_ind <- paste0(round((sum(df_sorted$draw_dev*100))/
                             (sum(abs((c(rep(settings_list$draws_number, settings_list$assembly_size), 
                                         rep(0, INPUT_SIZE - settings_list$assembly_size))) - draw_mean))),
                           2), '%')

# Summary statistics
summary_stats_data1 <- data.frame(settings_list$draws_number)
summary_stats_data1 <- append(summary_stats_data1, settings_list$SA_temperature)
summary_stats_data1 <- append(summary_stats_data1, settings_list$assembly_size)
summary_stats_data1 <- append(summary_stats_data1, INPUT_SIZE)
summary_stats1 <- data.frame(summary_stats_data1)
names(summary_stats1) <- c('Number of draws', 'Temperature', 'Size of the assembly', 'Number of volunteers')

summary_stats_data2 <- data.frame(round(exp(mean(log(results_summary$drawn_counter))),2))
summary_stats_data2 <- append(summary_stats_data2, round(sd(results_summary[,'drawn_counter']),2))
summary_stats_data2 <- append(summary_stats_data2, paste0(round(100 -length(which(results_summary$drawn_counter==0)) / 
                                                                  nrow(volunteers) * 100, 2), "%"))
summary_stats_data2 <- append(summary_stats_data2, settings_list$draws_number - sum(identical_columns))
summary_stats_data2 <- append(summary_stats_data2, paste0(gini_ind, "%"))
summary_stats_data2 <- append(summary_stats_data2, round(median(results_summary$drawn_counter),2))
summary_stats_data2 <- append(summary_stats_data2, tot_eq_ind)
summary_stats2 <- data.frame(summary_stats_data2)
names(summary_stats2) <- c('Geometric mean', 'Standard deviation', 'Inclusiveness', 'Unique panels', 
                           'Gini Index', 'Median', 'Total Equality Index')

summary_stats3 <- data.frame(round(draw_mean, 2))
names(summary_stats3) <- c('Equality')

# Add indicators worksheet
addWorksheet(wb, sheetName='Indicators')
writeDataTable(wb, sheet = 2, summary_stats1, startCol = 1, startRow = 1)
writeDataTable(wb, sheet = 2, summary_stats2, startCol = 1, startRow = 5)
writeDataTable(wb, sheet = 2, interval_eq, startCol = 1, startRow = 9)
writeDataTable(wb, sheet = 2, summary_stats3, startCol = 4, startRow = 9)
setColWidths(wb, sheet = 2, cols = 1:20, width = "auto")

results_summary <- results_summary[order(results_summary$ID),]

# Create histogram plot
histogram = ggplot(results_summary, aes(x = ID, y = drawn_counter)) + 
  geom_col(width = 0.1, colour="blue") + 
  theme(axis.text.x=element_blank()) + 
  xlab('ID') + 
  ylab('Number of selections')
plot_filename <- "plot.png"
png(plot_filename)
plot(histogram)
dev.off()

# Calculate accuracy index
accuracy_index = data.frame()
for (i in 1:settings_list$draws_number) {
  accuracy_index[1,i] = sum(abs(characteristics_summary$value - characteristics_summary[,6+i]))
}

# Calculate closeness index
closeness_index = data.frame()
for (i in 1:settings_list$draws_number) {
  closeness_index[1,i] = sum((abs(characteristics_summary$value - characteristics_summary[,6+i]))^1.6)
}

# Calculate duplicate count
duplicate_count <- data.frame()
for (i in 1:settings_list$draws_number) {
  duplicate_count <- rbind(duplicate_count, 
                           sum(results_summary[which(results_summary[,i+1] > 1), i+1]))
}
duplicate_count <- t(duplicate_count)

# Add composition overview worksheet
addWorksheet(wb, sheetName='Composition overview')
writeData(wb, sheet = 3, characteristics_summary[, colnames(characteristics_summary) != 'counter'])
writeData(wb, sheet = 3, 'Accuracy Index', startRow = categories_number + 3, startCol = 5)
writeData(wb, sheet = 3, 'Closeness Index', startRow = categories_number + 5, startCol = 5)
writeData(wb, sheet = 3, 'Duplicate count per volunteer ID', startRow = categories_number + 7, startCol = 5)
writeData(wb, sheet = 3, 'Draw time (in seconds)', startRow = categories_number + 9, startCol = 5)
writeData(wb, sheet = 3, 'Function value', startRow = categories_number + 11, startCol = 5)
writeData(wb, sheet = 3, 'Temperature', startRow = categories_number + 13, startCol = 5)
writeData(wb, sheet = 3, 'Average Accuracy Index', startRow = categories_number + 15, startCol = 5)
writeData(wb, sheet = 3, 'Average Closeness Index', startRow = categories_number + 17, startCol = 5)
writeData(wb, sheet = 3, 'Average Time', startRow = categories_number + 19, startCol = 5)
writeData(wb, sheet = 3, 'Median Time', startRow = categories_number + 21, startCol = 5)
writeData(wb, sheet = 3, 'Total drawing time', startRow = categories_number + 23, startCol = 5)

writeData(wb, sheet = 3, accuracy_index, startRow = categories_number + 3, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, closeness_index, startRow = categories_number + 5, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, duplicate_count, startRow = categories_number + 7, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, draws_times, startRow = categories_number + 9, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, function_value, startRow = categories_number + 11, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, settings_list$SA_temperature, startRow = categories_number + 13, startCol = 6)
writeData(wb, sheet = 3, rowMeans(accuracy_index), startRow = categories_number + 15, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, rowMeans(closeness_index), startRow = categories_number + 17, startCol = 6, 
          rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, round(mean(as.numeric(draws_times[1,])),2), startRow = categories_number + 19, 
          startCol = 6, rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, round(median(as.numeric(draws_times[1,])),2), startRow = categories_number + 21, 
          startCol = 6, rowNames=FALSE, colNames=FALSE)
writeData(wb, sheet = 3, round(as.numeric(difftime(end_draws_time, start_draws_time, units = "secs")),2), 
          startRow = categories_number + 23, startCol = 6, rowNames=FALSE, colNames=FALSE)
setColWidths(wb, sheet = 3, cols = 1:(5+settings_list$draws_number), width = "auto")

results_summary <- results_summary[order(results_summary$drawn_counter,decreasing = TRUE),]

sorted_results <- data.frame(cbind(results_summary$ID, results_summary$drawn_counter))
colnames(sorted_results) <- c('ID', 'Drawn Counter')

# Add sorted results worksheet
addWorksheet(wb, sheetName = "Sorted results")
writeDataTable(wb, sheet = 4, sorted_results, startRow = 1, startCol = 1)
setColWidths(wb, sheet = 4, cols = 1:2, width = "auto")

# Create bar plot
g3 <- ggplot(results_summary) + 
  geom_bar(aes(x= reorder(ID, drawn_counter), y = drawn_counter), 
           stat="identity", width=.7, position = "dodge", fill = '#0099FF') + 
  geom_line(aes(x = reorder(ID, drawn_counter), y = draw_mean), 
            group = 1, color='red', linewidth = 1) +
  coord_cartesian(ylim = c(0, settings_list$draws_number)) + 
  xlab('ID') + 
  ylab('Number of selections')

png(filename=paste0("GenSA_", settings_list$draws_number, "_barplot.png"), 
    width = 1700, height = 1300)
plot(g3)
dev.off()

# Create line plot
g4 <- ggplot(results_summary) + 
  geom_line(aes(x = reorder(ID, drawn_counter), y = drawn_counter), 
            group = 1, color='#0099FF', linewidth = 1) + 
  geom_line(aes(x = reorder(ID, drawn_counter), y = draw_mean), 
            group = 2, color='red', linewidth = 1) + 
  ylim(c(0, settings_list$draws_number)) + 
  xlab('ID') + 
  ylab('Number of selections')

png(filename=paste0("GenSA_", settings_list$draws_number, "_lineplot.png"), 
    width = 1700, height = 1300)
plot(g4)
dev.off()

# Save workbook and finish up
tryCatch({
  # Save workbook
  saveWorkbook(wb, xlsx_results_filename, overwrite = TRUE)
  
  console_and_file_print("Draw(s) completed")
  end_time_script <- Sys.time()
  script_time <- difftime(end_time_script, start_time_script, units = "secs")
  console_and_file_print(paste('Script running time:', round(as.numeric(script_time), 2), 'sec.'))
  
}, error = function(e) {
  console_and_file_print(paste("Error during final operations:", e$message))
}, finally = {
  # Sound notification if enabled
  directory_check <- getwd()
  if(!grepl("cloud", directory_check, fixed = TRUE) && settings_list$finish_signal) {
    tryCatch({
      beep(1)
    }, error = function(e) {
      warning("Could not play sound notification:", e$message)
    })
  }
  
  # Reset working directory
  setwd(main_directory)
})