  # ============================================================
  # Temperature setting for simulated annealing
  # of citizens' assemblies
  # Version 1.0
  # ============================================================
  
  calculate_sa_temperature <- function() {  
  library(openxlsx)
  
  # Add error handling for file loading
  tryCatch({
    settings <- read.xlsx("data_and_settings/settings.xlsx", sheet = 1, rowNames = TRUE)
    characteristics <- read.xlsx("data_and_settings/categories.xlsx", sheet = 1)
  }, error = function(e) {
    stop("Error loading files: ", e$message)
  })
  
  # Get assembly parameters
  assembly_size <- as.numeric(settings["assembly_size", "setting_value"])
  if(is.na(assembly_size) || assembly_size <= 0) {
    stop("Invalid assembly size: must be a positive number")
  }
  
  assembly_name <- if("draw_name" %in% rownames(settings)) {
    settings["draw_name", "setting_value"]
  } else {
    "Unnamed Assembly"
  }
  
  # Validate characteristics data structure
  if(nrow(characteristics) == 0) {
    stop("Categories file is empty")
  }
  
  required_columns <- c("category", "value")
  missing_columns <- setdiff(required_columns, names(characteristics))
  if(length(missing_columns) > 0) {
    stop("Missing required columns in categories file: ", paste(missing_columns, collapse = ", "))
  }
  
  # Clean data with checks
  characteristics_names <- lapply(names(characteristics), trimws)
  characteristics <- data.frame(lapply(characteristics, trimws))
  names(characteristics) <- characteristics_names
  
  # Check for empty categories
  empty_categories <- characteristics$category[characteristics$category == ""]
  if(length(empty_categories) > 0) {
    stop("Found empty category names")
  }
  
  # Basic structure analysis with validation
  unique_categories <- unique(characteristics$category)
  n_characteristics <- length(unique_categories)
  
  # Validate quotas for each characteristic
  cats_per_char <- lapply(unique_categories, function(cat) {
    subset <- characteristics[characteristics$category == cat, ]
    values <- as.numeric(subset$value)
    
    # Validate numeric values
    if(any(is.na(values))) {
      stop("Invalid numeric values found in category: ", cat)
    }
    
    # Validate non-negative values
    if(any(values < 0)) {
      stop("Negative quota values found in category: ", cat)
    }
    
    # Validate quotas don't exceed assembly size
    if(any(values > assembly_size)) {
      stop("Quota value exceeds assembly size in category: ", cat)
    }
    
    # Check if category has any members
    if(sum(values) == 0) {
      stop("Category has no members: ", cat)
    }
    
    # Validate total matches assembly size (with small tolerance for rounding)
    total <- sum(values)
    if(abs(total - assembly_size) > 0.01) {
      stop(sprintf("Total quota (%0.2f) doesn't match assembly size (%d) for category: %s", 
                   total, assembly_size, cat))
    }
    
    avg_quota <- mean(values)
    cv <- sd(values) / avg_quota
    min_quota <- min(values)
    
    list(
      count = length(values),
      values = values,
      cv = cv,
      min_quota = min_quota,
      quota_ratio = min_quota / assembly_size
    )
  })
  
  # Calculate structural metrics with validation
  max_cats <- max(sapply(cats_per_char, function(x) x$count))
  total_cats <- sum(sapply(cats_per_char, function(x) x$count))
  
  # Base temperature calculation with validation
  base_temp <- if(n_characteristics <= 3 && max_cats <= 6) {
    temp <- 500 * (total_cats / assembly_size)^0.35
    temp
  } else {
    temp <- 600 * (total_cats / assembly_size)^0.4
    temp
  }
  
  # Geographic complexity with validation
  geo_complexity <- if(max_cats > 15) {
    2000 * log10(max_cats)^2
  } else if(max_cats > 8) {
    1000 * log10(max_cats)^1.5
  } else {
    175 * sqrt(max_cats)
  }
  
  # Quota difficulty with validation
  quota_difficulty <- mean(sapply(cats_per_char, function(x) {
    if(x$count > 1) {
      diff_factor <- x$cv * sqrt(x$count)
      if(x$count > 12) {
        diff_factor * 1500
      } else if(x$count > 6) {
        diff_factor * 750
      } else {
        diff_factor * 150
      }
    } else {
      0
    }
  }))
  
  # Interaction base with validation - modified for >7 characteristics
  interaction_base <- if(n_characteristics <= 7) {
    switch(n_characteristics,
           0,     # 1 characteristic
           150,   # 2 characteristics
           250,   # 3 characteristics
           750,   # 4 characteristics
           1400,  # 5 characteristics
           2000,  # 6 characteristics
           2600   # 7 characteristics
    )
  } else {
    2600 + (n_characteristics - 7) * 400  # Scale up for more than 7 characteristics
  }
  
  # Interaction multiplier with validation
  interaction_multiplier <- if(n_characteristics >= 6) {
    1.6 + (n_characteristics - 5) * log10(total_cats/n_characteristics)
  } else if(n_characteristics >= 4) {
    1.2 + (n_characteristics - 3) * 0.1
  } else {
    0.9
  }
  
  # Size factor with validation
  size_factor <- if(n_characteristics <= 3 && max_cats <= 6) {
    1 + (50 - assembly_size) / 200
  } else {
    1 + (50 - assembly_size) / 150
  }
  
  # Complexity score calculation
  complexity_score <- base_temp + geo_complexity + quota_difficulty
  
  # Apply selective multiplication
  if(max_cats > 15) {
    complexity_score <- complexity_score * 1.8
  } else if(max_cats > 8) {
    complexity_score <- complexity_score * 1.3
  }
  
  # Final temperature calculation with validation
  final_temp <- (complexity_score + interaction_base * interaction_multiplier) * size_factor
  
  if(is.na(final_temp) || is.null(final_temp)) {
    stop("Final temperature calculation resulted in NA or NULL")
  }
  
  # Additional scaling for complex cases
  if(n_characteristics >= 6 && max_cats > 5) {
    final_temp <- final_temp * 1.45
  }
  
  # Rounding with validation
  rounded_temp <- if(final_temp < 2000) {
    round(final_temp / 100) * 100
  } else if(final_temp < 5000) {
    round(final_temp / 250) * 250
  } else {
    round(final_temp / 1000) * 1000
  }
  
  # Print analysis to console
  cat("\nTemperature Analysis for:", assembly_name, "\n")
  cat("----------------------------------------\n")
  cat(sprintf("Assembly size: %d\n", assembly_size))
  cat(sprintf("Number of characteristics: %d\n", n_characteristics))
  cat(sprintf("Total categories: %d\n", total_cats))
  cat(sprintf("Maximum categories in characteristic: %d\n", max_cats))
  cat(sprintf("Base temperature: %.0f\n", base_temp))
  cat(sprintf("Geographic complexity: %.0f\n", geo_complexity))
  cat(sprintf("Quota difficulty: %.0f\n", quota_difficulty))
  cat(sprintf("Interaction base: %.0f\n", interaction_base))
  cat(sprintf("Interaction multiplier: %.2f\n", interaction_multiplier))
  cat(sprintf("Complexity score: %.0f\n", complexity_score))
  cat(sprintf("Size factor: %.2f\n", size_factor))
  cat("----------------------------------------\n")
  cat(sprintf("Recommended temperature: %d\n", rounded_temp))
  
  return(rounded_temp)
}

# Run calculation with error handling
tryCatch({
  recommended_temp <- calculate_sa_temperature()
}, error = function(e) {
  cat("Error occurred:", e$message, "\n")
})